@@ -8,5 +8,7 @@ DEFAULT_ROW_COUNT = 10
 DEFAULT_COLUMN_COUNT = 5
 EPOCH = datetime(2001, 1, 1)
 MAX_TILE_SIZE = 256
+MAX_ROW_COUNT = 1000000
+MAX_COL_COUNT = 1000
 DOCUMENT_ID = 1
 PACKAGE_ID = 2
