@@ -281,11 +281,7 @@ def get_archive_info_and_remainder(buf):
 
 def create_iwa_segment(id: int, cls: object, object_dict: dict) -> object:
     full_name = cls.DESCRIPTOR.full_name
-    # TODO: fix the ID mapping hack
-    if full_name == "TST.TableDataList":
-        type_id = 6005
-    else:
-        type_id = NAME_ID_MAP[full_name]
+    type_id = NAME_ID_MAP[full_name]
     header = {
         "_pbtype": "TSP.ArchiveInfo",
         "identifier": str(id),
@@ -332,18 +328,3 @@ def copy_object_to_iwa_file(iwa_file: IWAFile, obj: object, obj_id: int):
                     _ = msg_info.object_references.pop()
                 for reference in references:
                     msg_info.object_references.append(reference)
-
-
-def deep_print(obj, indent=0):
-    for descriptor in obj.DESCRIPTOR.fields:
-        value = getattr(obj, descriptor.name)
-        if descriptor.type == descriptor.TYPE_MESSAGE:
-            if descriptor.label == descriptor.LABEL_REPEATED:
-                map(lambda x: deep_print(x, indent + 1), value)
-            else:
-                deep_print(value, indent + 1)
-        elif descriptor.type == descriptor.TYPE_ENUM:
-            enum_name = descriptor.enum_type.values[value].name
-            print("  " * indent + f"{descriptor.full_name}, {enum_name}")
-        else:
-            print("  " * indent + f"{descriptor.full_name} {value}")
