@@ -40,7 +40,6 @@ from numbers_parser.bullets import (
 )
 from numbers_parser.generated import TNArchives_pb2 as TNArchives
 from numbers_parser.generated import TSDArchives_pb2 as TSDArchives
-from numbers_parser.generated import TSKArchives_pb2 as TSKArchives
 from numbers_parser.generated import TSPMessages_pb2 as TSPMessages
 from numbers_parser.generated import TSPArchiveMessages_pb2 as TSPArchiveMessages
 from numbers_parser.generated import TSTArchives_pb2 as TSTArchives
@@ -337,13 +336,22 @@ class _NumbersModel:
         return cell_records
 
     @lru_cache(maxsize=None)
-    def calc_engine(self):
-        """Return the CalculationEngine object for the current document"""
+    def calc_engine_id(self):
+        """Return the CalculationEngine ID for the current document"""
         ce_id = self.find_refs("CalculationEngineArchive")
         if len(ce_id) == 0:
+            return 0
+        else:
+            return ce_id[0]
+
+    @lru_cache(maxsize=None)
+    def calc_engine(self):
+        """Return the CalculationEngine object for the current document"""
+        ce_id = self.calc_engine_id()
+        if ce_id == 0:
             return None
         else:
-            return self.objects[ce_id[0]]
+            return self.objects[ce_id]
 
     def calculate_merge_cell_ranges(self, table_id):
         """Exract all the merge cell ranges for the Table."""
@@ -567,33 +575,9 @@ class _NumbersModel:
                 merge_map.cell_range.append(cell_range)
 
         base_data_store = self.objects[table_id].base_data_store
-        base_data_store.merge_region_map.MergeFrom(
+        base_data_store.merge_region_map.CopyFrom(
             TSPMessages.Reference(identifier=merge_map_id)
         )
-        # TODO deal with Assertion:
-        # Assert *** Assertion failure #0: -[TSPUnarchiver
-        #   validateReferenceToObjectIdentifier:objectClass:isWeak
-        #   :validateStrongReferences:selector:weakSelector:]
-        #   /Library/Caches/com.apple.xbs/Sources/iWorkDependenciesMacOS/
-        #   iWorkDependenciesMacOS-7034.0.86/shared/persistence/src/TSPUnarchiver.mm:704
-        #   Object [TSTMergeRegionMap-6347027] is not strongly referenced from message
-        #   TST.TableModelArchive-1058328] message type 6001.
-        #   Use TSPReadWeakReferenceMessage<TSTMergeRegionMap>(unarchiver, message,
-        #   completion) instead.
-
-        # component_map = {c.identifier: c for c in self.objects[PACKAGE_ID].components}
-        # component_id = [
-        #     id
-        #     for id, c in component_map.items()
-        #     if c.preferred_locator == "CalculationEngine"
-        # ][0]
-        # component_map[component_id].external_references.append(
-        #     TSPArchiveMessages.ComponentExternalReference(
-        #         component_identifier=table_id,
-        #         object_identifier=merge_map_id,
-        #         is_weak=True,
-        #     )
-        # )
 
     def recalculate_column_row_uid_map(self, table_id: int, data: List):
         table_model = self.objects[table_id]
@@ -674,45 +658,53 @@ class _NumbersModel:
         row_info.has_wide_offsets = wide_offsets
         return row_info
 
-    def update_package_metadata(self, obj_id: int, parent: str, locator: str = None):
+    @lru_cache(maxsize=None)
+    def metadata_component(self, component_name: str) -> int:
+        """Return the ID of an object in the document metadata given it's name"""
         component_map = {c.identifier: c for c in self.objects[PACKAGE_ID].components}
-        if obj_id not in component_map:
-            component_id = [
-                id for id, c in component_map.items() if c.preferred_locator == parent
-            ][0]
+        component_ids = [
+            id
+            for id, c in component_map.items()
+            if c.preferred_locator == component_name
+        ]
+        return component_map[component_ids[0]]
 
-            if component_id is not None:
-                if locator is not None:
-                    locator = locator.format(obj_id)
-                    preferred_locator = re.sub(r"\-\d+.*", "", locator)
-                    component_info = TSPArchiveMessages.ComponentInfo(
-                        identifier=obj_id,
-                        locator=locator,
-                        preferred_locator=preferred_locator,
-                        is_stored_outside_object_archive=False,
-                        save_token=1,
-                    )
-                else:
-                    component_info = TSPArchiveMessages.ComponentInfo(
-                        identifier=obj_id,
-                        preferred_locator=parent,
-                        save_token=1,
-                    )
+    def add_component_metadata(self, object_id: int, parent: str, locator: str):
+        """Add a new ComponentInfo record to the parent object in the document metadata"""
+        locator = locator.format(object_id)
+        preferred_locator = re.sub(r"\-\d+.*", "", locator)
+        component_info = TSPArchiveMessages.ComponentInfo(
+            identifier=object_id,
+            locator=locator,
+            preferred_locator=preferred_locator,
+            is_stored_outside_object_archive=False,
+            document_read_version=[2, 0, 0],
+            document_write_version=[2, 0, 0],
+            save_token=1,
+        )
+        self.objects[PACKAGE_ID].components.append(component_info)
+        self.add_component_reference(object_id, parent)
+        self.objects.mark_as_dirty(PACKAGE_ID)
 
-                self.objects[PACKAGE_ID].components.append(component_info)
-                component_map[component_id].external_references.append(
-                    TSPArchiveMessages.ComponentExternalReference(
-                        component_identifier=obj_id
-                    )
-                )
-            else:
-                component_map[component_id].external_references.append(
-                    TSPArchiveMessages.ComponentExternalReference(
-                        component_identifier=component_id, object_identifier=obj_id
-                    )
-                )
-
-            self.objects.mark_as_dirty(PACKAGE_ID)
+    def add_component_reference(
+        self,
+        object_id: int,
+        location: str,
+        parent_id: int = None,
+        is_weak: bool = False,
+    ):
+        """Add an external reference to an object in a metadata component"""
+        component = self.metadata_component(location)
+        if parent_id is not None:
+            params = {"object_identifier": object_id, "component_identifier": parent_id}
+        else:
+            params = {"component_identifier": object_id}
+        if is_weak:
+            params["is_weak"] = True
+        component.external_references.append(
+            TSPArchiveMessages.ComponentExternalReference(**params)
+        )
+        self.objects.mark_as_dirty(PACKAGE_ID)
 
     def recalculate_table_data(self, table_id: int, data: List):
         table_model = self.objects[table_id]
@@ -730,7 +722,7 @@ class _NumbersModel:
         tile_idx = 0
         max_tile_idx = len(data) >> 8
         base_data_store = self.objects[table_id].base_data_store
-        tile_ids = [t.tile.identifier for t in base_data_store.tiles.tiles]
+        base_data_store.tiles.ClearField("tiles")
         if len(data[0]) > MAX_TILE_SIZE:
             base_data_store.tiles.should_use_wide_rows = True
 
@@ -743,61 +735,30 @@ class _NumbersModel:
                 num_rows = len(data) - row_start
                 row_end = row_start + num_rows
 
-            if tile_idx > (len(tile_ids) - 1):
-                tile_dict = {
-                    "maxColumn": 0,
-                    "maxRow": 0,
-                    "numCells": 0,
-                    "numrows": num_rows,
-                    "storage_version": 5,
-                    "rowInfos": [],
-                    "last_saved_in_BNC": True,
-                    "should_use_wide_rows": True,
-                }
-                tile_id, tile = self.objects.create_object_from_dict(
-                    "Index/Tables/Tile-{}", tile_dict, TSTArchives.Tile
-                )
-                for row_num in range(row_start, row_end):
-                    row_info = self.recalculate_row_info(
-                        table_id, data, row_start, row_num
-                    )
-                    tile.rowInfos.append(row_info)
+            tile_dict = {
+                "maxColumn": 0,
+                "maxRow": 0,
+                "numCells": 0,
+                "numrows": num_rows,
+                "storage_version": 5,
+                "rowInfos": [],
+                "last_saved_in_BNC": True,
+                "should_use_wide_rows": True,
+            }
+            tile_id, tile = self.objects.create_object_from_dict(
+                "Index/Tables/Tile-{}", tile_dict, TSTArchives.Tile
+            )
+            for row_num in range(row_start, row_end):
+                row_info = self.recalculate_row_info(table_id, data, row_start, row_num)
+                tile.rowInfos.append(row_info)
 
-                tile_ref = TSTArchives.TileStorage.Tile()
-                tile_ref.tileid = tile_idx
-                tile_ref.tile.MergeFrom(TSPMessages.Reference(identifier=tile_id))
-                base_data_store.tiles.tiles.append(tile_ref)
-                base_data_store.tiles.tile_size = MAX_TILE_SIZE
+            tile_ref = TSTArchives.TileStorage.Tile()
+            tile_ref.tileid = tile_idx
+            tile_ref.tile.MergeFrom(TSPMessages.Reference(identifier=tile_id))
+            base_data_store.tiles.tiles.append(tile_ref)
+            base_data_store.tiles.tile_size = MAX_TILE_SIZE
 
-                self.update_package_metadata(
-                    tile_id, "CalculationEngine", "Tables/Tile-{}"
-                )
-
-                # TODO: is this required?
-                base_data_store.rowTileTree.nodes.append(
-                    TSTArchives.TableRBTree.Node(key=row_start, value=tile_idx)
-                )
-            else:
-                tile_id = tile_ids[tile_idx]
-                tile = self.objects[tile_id]
-                tile.maxColumn = 0
-                tile.maxRow = 0
-                tile.numCells = 0
-                tile.numrows = num_rows
-                tile.storage_version = 5
-                tile.last_saved_in_BNC = True
-                tile.should_use_wide_rows = True
-                clear_field_container(tile.rowInfos)
-
-                for row_num in range(row_start, row_end):
-                    row_info = self.recalculate_row_info(
-                        table_id, data, row_start, row_num
-                    )
-                    tile.rowInfos.append(row_info)
-
-                self.update_package_metadata(
-                    tile_id, "CalculationEngine", "Tables/Tile-{}"
-                )
+            self.add_component_metadata(tile_id, "CalculationEngine", "Tables/Tile-{}")
 
             tile_idx += 1
 
@@ -809,7 +770,7 @@ class _NumbersModel:
             {"listType": TSTArchives.TableDataList.ListType.STRING, "nextListID": 1},
             TSTArchives.TableDataList,
         )
-        self.update_package_metadata(
+        self.add_component_metadata(
             table_strings_id, "CalculationEngine", "Tables/DataList-{}"
         )
         return table_strings_id, table_strings
@@ -823,16 +784,39 @@ class _NumbersModel:
         }
         return style_sheet_map
 
-    def add_table(self, sheet_id: int, table_name: str) -> int:
-        from_table_id = self.table_ids(sheet_id)[-1]
-        from_table = self.objects[from_table_id]
+    def sheet_table_offset(self, sheet_id):
+        """Sum the heights of all rows in all tables in a sheet"""
+        if len(self.table_ids(sheet_id)) == 0:
+            return 0.0
 
-        print("\n")
+        height = 0.0
+        for table_id in self.table_ids(sheet_id):
+            table_model = self.objects[table_id]
+            # height = table_model.default_row_height * number_of_rows
+
+            bds = self.objects[table_id].base_data_store
+            buckets = self.objects[bds.rowHeaders.buckets[0].identifier].headers
+            for i, row in self.row_storage_map(table_id).items():
+                if row is not None:
+                    height += buckets[i].size
+                else:
+                    height += table_model.default_row_height
+
+        # TODO: calculate gap between tables
+        height += 100.0
+
+        return height
+
+    def add_table(
+        self, sheet_id: int, table_name: str, from_table_id: int = None
+    ) -> int:
+        if from_table_id is None:
+            from_table_id = self.table_ids(sheet_id)[-1]
+        from_table = self.objects[from_table_id]
 
         table_strings_id, table_strings = self.create_string_table()
 
-        print(f"table_strings_id = {table_strings_id}")
-
+        from_table_refs = field_references(from_table)
         table_model_id, table_model = self.objects.create_object_from_dict(
             "CalculationEngine",
             {
@@ -840,47 +824,20 @@ class _NumbersModel:
                 "number_of_rows": DEFAULT_ROW_COUNT,
                 "number_of_columns": DEFAULT_COLUMN_COUNT,
                 "table_name": table_name,
+                "table_name_enabled": True,
                 "default_row_height": 20.0,
                 "default_column_width": 98.0,
-                "table_style": {"identifier": from_table.table_style.identifier},
-                "body_text_style": {
-                    "identifier": from_table.body_text_style.identifier
-                },
-                "header_row_text_style": {
-                    "identifier": from_table.header_row_text_style.identifier
-                },
-                "header_column_text_style": {
-                    "identifier": from_table.header_column_text_style.identifier
-                },
-                "footer_row_text_style": {
-                    "identifier": from_table.footer_row_text_style.identifier
-                },
-                "body_cell_style": {
-                    "identifier": from_table.body_cell_style.identifier
-                },
-                "header_row_style": {
-                    "identifier": from_table.header_row_style.identifier
-                },
-                "header_column_style": {
-                    "identifier": from_table.header_column_style.identifier
-                },
-                "footer_row_style": {
-                    "identifier": from_table.footer_row_style.identifier
-                },
+                **from_table_refs,
             },
             TSTArchives.TableModelArchive,
         )
 
-        print(f"table_model_id = {table_model_id}")
-
         column_headers_id, column_headers = self.objects.create_object_from_dict(
             "Index/Tables/HeaderStorageBucket-{}",
             {"bucketHashFunction": 1},
             TSTArchives.HeaderStorageBucket,
         )
-        print(f"column_headers_id = {column_headers_id}")
-
-        self.update_package_metadata(
+        self.add_component_metadata(
             column_headers_id, "CalculationEngine", "Tables/HeaderStorageBucket-{}"
         )
 
@@ -889,9 +846,8 @@ class _NumbersModel:
             {"listType": TSTArchives.TableDataList.ListType.STYLE, "nextListID": 1},
             TSTArchives.TableDataList,
         )
-        print(f"style_table_id = {style_table_id}")
 
-        self.update_package_metadata(
+        self.add_component_metadata(
             style_table_id, "CalculationEngine", "Tables/DataList-{}"
         )
 
@@ -900,9 +856,8 @@ class _NumbersModel:
             {"listType": TSTArchives.TableDataList.ListType.FORMULA, "nextListID": 1},
             TSTArchives.TableDataList,
         )
-        print(f"formula_table_id = {formula_table_id}")
 
-        self.update_package_metadata(
+        self.add_component_metadata(
             formula_table_id, "CalculationEngine", "Tables/TableDataList-{}"
         )
 
@@ -911,29 +866,30 @@ class _NumbersModel:
             {"listType": TSTArchives.TableDataList.ListType.STYLE, "nextListID": 1},
             TSTArchives.TableDataList,
         )
-        print(f"format_table_pre_bnc_id = {format_table_pre_bnc_id}")
 
-        self.update_package_metadata(
+        self.add_component_metadata(
             format_table_pre_bnc_id,
-            "DocumentStylesheet",
+            "Document",
             "Tables/TableDataList-{}",
         )
 
+        data_store_refs = field_references(from_table.base_data_store)
+        data_store_refs["stringTable"] = {"identifier": table_strings_id}
+        data_store_refs["columnHeaders"] = {"identifier": column_headers_id}
+        data_store_refs["styleTable"] = {"identifier": style_table_id}
+        data_store_refs["formula_table"] = {"identifier": formula_table_id}
+        data_store_refs["format_table_pre_bnc"] = {
+            "identifier": format_table_pre_bnc_id
+        }
         table_model.base_data_store.MergeFrom(
             TSTArchives.DataStore(
-                stringTable=TSPMessages.Reference(identifier=table_strings_id),
                 rowHeaders=TSTArchives.HeaderStorage(bucketHashFunction=1),
-                columnHeaders=TSPMessages.Reference(identifier=column_headers_id),
                 nextRowStripID=1,
                 nextColumnStripID=0,
                 rowTileTree=TSTArchives.TableRBTree(),
                 columnTileTree=TSTArchives.TableRBTree(),
                 tiles=TSTArchives.TileStorage(tile_size=256, should_use_wide_rows=True),
-                styleTable=TSPMessages.Reference(identifier=style_table_id),
-                formula_table=TSPMessages.Reference(identifier=formula_table_id),
-                format_table_pre_bnc=TSPMessages.Reference(
-                    identifier=format_table_pre_bnc_id
-                ),
+                **data_store_refs,
             )
         )
 
@@ -950,9 +906,8 @@ class _NumbersModel:
             {"bucketHashFunction": 1},
             TSTArchives.HeaderStorageBucket,
         )
-        print(f"row_headers_id = {row_headers_id}")
 
-        self.update_package_metadata(
+        self.add_component_metadata(
             row_headers_id, "CalculationEngine", "Tables/HeaderStorageBucket-{}"
         )
         table_model.base_data_store.rowHeaders.buckets.append(
@@ -961,14 +916,12 @@ class _NumbersModel:
 
         self.recalculate_table_data(table_model_id, data)
 
-        # TODO: fix package metadata references
-
         table_info_id, table_info = self.objects.create_object_from_dict(
             "CalculationEngine",
             {},
             TSTArchives.TableInfoArchive,
         )
-        print(f"table_info_id = {table_info_id}")
+        self.add_component_reference(table_info_id, "Document", self.calc_engine_id())
 
         table_info.tableModel.MergeFrom(
             TSPMessages.Reference(identifier=table_model_id)
@@ -978,8 +931,8 @@ class _NumbersModel:
             geometry=TSDArchives.GeometryArchive(
                 angle=0.0,
                 flags=3,
-                position=TSPMessages.Point(x=0.0, y=400.0),
-                size=TSPMessages.Size(height=200.0, width=500.0),
+                position=TSPMessages.Point(x=0.0, y=self.sheet_table_offset(sheet_id)),
+                size=TSPMessages.Size(height=231.0, width=494.0),
             ),
         )
         table_info.super.MergeFrom(drawable)
@@ -988,34 +941,19 @@ class _NumbersModel:
             TSPMessages.Reference(identifier=table_info_id)
         )
 
-        # TODO (1): is this required?
-        new_tree_node_id, tree_node = self.objects.create_object_from_dict(
-            "Document",
-            {},
-            TSKArchives.TreeNode,
-        )
-        tree_node.object.MergeFrom(TSPMessages.Reference(identifier=table_info_id))
-        for tree_node_id in self.objects.find_refs("TreeNode"):
-            if self.objects[tree_node_id].object.identifier == sheet_id:
-                self.objects[tree_node_id].children.append(
-                    TSPMessages.Reference(identifier=new_tree_node_id)
-                )
-        # TODO (2): is this required
-        self.update_package_metadata(table_info_id, "CalculationEngine")
-
         return table_model_id
 
-    def add_sheet(self, sheet_name: str, table_name: str, from_sheet: int):
+    def add_sheet(self, sheet_name: str, table_name: str, from_table_id: int):
+        """Add a new sheet with a copy of a table from another sheet"""
         sheet_id, sheet_archive = self.objects.create_object_from_dict(
             "Document", {"name": sheet_name}, TNArchives.SheetArchive
         )
 
-        table_info_id = self.add_table(from_sheet, table_name)
+        table_info_id = self.add_table(sheet_id, table_name, from_table_id)
 
         self.objects[DOCUMENT_ID].sheets.append(
             TSPMessages.Reference(identifier=sheet_id)
         )
-
         sheet_archive.drawable_infos.append(
             TSPMessages.Reference(identifier=table_info_id)
         )
@@ -1212,7 +1150,7 @@ class _NumbersModel:
 
         return cell_value
 
-    def duration_format(
+    def duration_format(  # noqa: C901
         self,
         cell_value: int,
         format_id: int,
@@ -1644,3 +1582,13 @@ def auto_units(cell_value, format):
             unit_smallest = unit_largest
 
     return unit_smallest, unit_largest
+
+
+def field_references(obj: object) -> dict:
+    """Return a dict of all fields in an object that are references to other objects"""
+    refs = {
+        x[0].name: {"identifier": getattr(obj, x[0].name).identifier}
+        for x in obj.ListFields()
+        if type(getattr(obj, x[0].name)) == TSPMessages.Reference
+    }
+    return refs
